package geoinformatics1;

import static geoinformatics1.CellState.*;

import java.util.*;

import repast.simphony.context.Context;
import repast.simphony.engine.schedule.*;
import repast.simphony.query.space.grid.*;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.*;
import repast.simphony.util.ContextUtils;

public class Cell
{
	public int getAge()
	{
		return age;
	}

	//The current state of the cell
	CellState state = EMPTY;
	//The updated state of the cell, which will be applied once all the cells have calculated their next states
	CellState nextState = EMPTY;
	//The grid on which this cell is located - used to find neighbours etc.
	Grid<Cell> grid = null;
	//The cell's Context
	Context<Cell> context = null;
	//A list of the cell's neighbours, currently the Moore neighbourhood, i.e. 8 cells
	List<Cell> neighbours = null;
	//The number of ticks the cell has spent in the current state
	int age = 0;
	//The probability of empty land with one or more developed neighbours being released for development
	double releaseProbability = 0.3;
	//The probability of released land being built on
	double buildProbability = 0.5;
	//The probability of old development (i.e. age > REDEVELOPMENT_AGE) being redeveloped
	double redevelopmentProbability = 1.0;
	//The age after which cells are considered for redevelopment
	public static final int REDEVELOPMENT_AGE = 100;

	/**
	 * Main update method: cell behaviour goes in here - use the different cases in the switch
	 * statement to specify rules for when the cell is in each state. Use setState( OLD|NEW etc. ) to
	 * set the new state for the cell.
	 * 
	 * Hint: to perform probabilistic operations, you can use: if( RandomHelper.nextDouble() < prob ) ...
	 * See also: hasNeighbour() and isSurrounded(), and later functions in this file.
	 * Hint: all of the behaviour can be done with just a line or two per case
	 */
	@ScheduledMethod(start = 1, interval = 1, priority = 1)
	public void update()
	{
		preUpdate();
		age++;

		switch ( state )
		{
		case EMPTY:
			//Behaviour for empty cells
			if( hasNeighbour() && RandomHelper.nextDouble() < releaseProbability ) setState( AVAILABLE );
			break;
		case AVAILABLE:
			//Behaviour for available cells
			if (RandomHelper.nextDouble() < buildProbability ) setState(NEW);
			break;
		case NEW:
			//Behaviour for new cells
			if( isSurrounded() ) setState(OLD);
			break;
		case OLD:
			//Behaviour for old cells
			if (age >= REDEVELOPMENT_AGE && RandomHelper.nextDouble() < redevelopmentProbability) setState(AVAILABLE);
			//if (age >= REDEVELOPMENT_AGE) setState(AVAILABLE);
			break;
		}
	}
	
	/*
	 * Functions after here are useful, but should not need changing
	 */
	
	/**
	 * Use this to set the state of the cell to a new state (e.g. setState(NEW) )
	 * @param state
	 */
	public void setState( CellState state )
	{
		nextState = state;
	}
	
	
	
	/**
	 * Returns true if the cell is surrounded by NEW or OLD development
	 */
	boolean isSurrounded()
	{
		return getDevelopedNeighbours() == neighbours.size();
	}
	
	/**
	 * Return true if the cell has at least one neighbour which is developed (NEW or OLD)
	 * @return
	 */
	public boolean hasNeighbour()
	{
		return getDevelopedNeighbours() > 0;
	}
	
	/**
	 * Returns the number of NEW or OLD neighbours the cell has
	 * @return
	 */
	public int getDevelopedNeighbours()
	{
		int num = 0;
		for( Cell c : neighbours ) if( c.getState().equals( NEW ) || c.getState().equals( OLD ) ) num++;
		return num;
	}
	
	
	/*
	 * Functions after here are unlikely to be useful
	 */
	
	@ScheduledMethod(start=1, interval=1, priority=ScheduleParameters.LAST_PRIORITY )
	public void updateState()
	{
		if( nextState != state ) age = 0;
		state = nextState;
	}
	
	@SuppressWarnings("unchecked")
	protected void preUpdate()
	{
		if( context == null )
			context = (Context<Cell>) ContextUtils.getContext( this );

		if( grid == null )
			grid = (Grid<Cell>) context.getProjection( "Grid" );

		if( neighbours == null )
		{
			MooreQuery<Cell> query = new MooreQuery<Cell>( grid, this );
			neighbours = new ArrayList<Cell>();
			for ( Cell c : query.query() )
				neighbours.add( c );
		}
	}


	public CellState getState()
	{
		return state;
	}
	
	public String getStateString()
	{
		return state.toString();
	}
	
	public boolean isEmpty() { return CellState.EMPTY.equals( state ); }
	public boolean isAvailable() { return CellState.AVAILABLE.equals( state ); }
	public boolean isNew() { return CellState.NEW.equals( state ); }
	public boolean isOld() { return CellState.OLD.equals( state ); }
}
