package geoinformatics3.city;

import static geoinformatics3.city.CellState.*;
import static java.lang.Math.*;

import geoinformatics3.*;
import geoinformatics3.agents.*;

import java.util.*;

import repast.simphony.context.Context;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.schedule.*;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.space.grid.*;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.*;
import repast.simphony.util.ContextUtils;
import sun.nio.cs.Surrogate;

public class Cell
{
	public int getAge()
	{
		return age;
	}
	
	CombinedModel model;
	City city;

	//The current state of the cell
	CellState state = EMPTY;
	//The updated state of the cell, which will be applied once all the cells have calculated their next states
	CellState nextState = EMPTY;
	//The grid on which this cell is located - used to find neighbours etc.
	Grid<Cell> grid = null;
	//The cell's Context
	Context<Cell> context = null;
	//A list of the cell's neighbours, currently the Moore neighbourhood, i.e. 8 cells
	List<Cell> neighbours = null;
	//The number of ticks the cell has spent in the current state
	int age = 0;
	//The probability of empty land with one or more developed neighbours being released for development
	public static double RELEASE_PROBABILITY = 0.1;
	//The probability of old development (i.e. age > REDEVELOPMENT_AGE) being redeveloped
	public static double REDEVELOPMENT_RATE = 0.2;
	//The age after which cells are considered for redevelopment
	public static int REDEVELOPMENT_AGE = 30;
	//The proportion of housing to allow to be surplus
	public static double SURPLUS_FACTOR = 0.1;
	
	
	//The current capacity of the cell
	public static int MAX_CAPACITY = 1;
	public static int INITIAL_CAPACITY = 1;
	public static int CAPACITY_INCREMENT = 1;
	public static int CURRENT_MAXIMUM_CAPACITY = INITIAL_CAPACITY;
	int capacity = INITIAL_CAPACITY;
	//The agents who are living in the cell
	Set<Agent> agents = new HashSet<Agent>(10);
	
	public Cell( CombinedModel model )
	{
		this.model = model;
		this.city = model.getCity();
	}
	
	/**
	 * Main update method: cell behaviour goes in here - use the different cases in the switch
	 * statement to specify rules for when the cell is in each state. Use setState( OLD|NEW etc. ) to
	 * set the new state for the cell.
	 * 
	 * Hint: to perform probabilistic operations, you can use: if( RandomHelper.nextDouble() < prob ) ...
	 * See also: hasNeighbour() and isSurrounded(), and later functions in this file.
	 * Hint: all of the behaviour can be done with just a line or two per cases
	 * 
	 * Hint: you can use city.getTotalCapacity() and city.getTotalPopulation() for capacity and population
	 * 
	 * Hint: Use SURPLUS_FACTOR for the surplus.
	 * 
	 * Hint: now make sure that you use build() and redevelop() instead of setState( NEW ) and setState( AVAILABLE )
	 * respectively, to make sure the correct things happen to the cell.
	 */
	@ScheduledMethod( start = 1, interval=1, priority=10 )
	public void update()
	{
		preUpdate();
		age++;

		switch ( state )
		{
		case EMPTY:
			if( hasNeighbour() && RandomHelper.nextDouble() < 0.1 ) setState( AVAILABLE );
			break;
		case AVAILABLE:
			if( RandomHelper.nextDouble() < 0.2 ) setState( NEW );
			break;
		case NEW:
			if( isSurrounded() ) setState( OLD );
			break;
		case OLD:
			if( age > REDEVELOPMENT_AGE && RandomHelper.nextDouble() < 0.3 ) setState( AVAILABLE );
			break;
		}
	}
	
	/**
	 * Use this function to redevelop cells; it evicts all the residents, increases the capacity
	 * and rebuilds
	 */
	void redevelop()
	{
		evictResidents();
		model.getCity().addHouses( -capacity );
		incrementCapacity();
		build();
	}
	
	/**
	 * Use this function to transition from AVAILABLE to NEW - it updates the city about the
	 * current housing capacity
	 */
	void build()
	{
		setState( NEW );
		model.getCity().addHouses( capacity );
	}
	
	public void incrementCapacity()
	{
		capacity += CAPACITY_INCREMENT;
		if( capacity > MAX_CAPACITY ) capacity = MAX_CAPACITY;
		if( capacity > CURRENT_MAXIMUM_CAPACITY ) CURRENT_MAXIMUM_CAPACITY = capacity;
	}
	
	/*
	 * Functions after here are useful, but should not need changing
	 */
	
	/**
	 * Use this to set the state of the cell to a new state (e.g. setState(NEW) )
	 * @param state
	 */
	public void setState( CellState state )
	{
		nextState = state;
	}
	
	/**
	 * Returns the average ratio of population to capcity in the neighbourhood
	 * @return
	 */
	public double getNeighbourhoodCapacityRatio()
	{
		if( neighbours == null ) return 0;
		double fullness = 0;
		for( Cell c : neighbours ) fullness += c.getCapacityRatio();
		fullness += getCapacityRatio();
		return fullness / (neighbours.size() + 1);
	}
	
	public double getNeighbourhoodDensity()
	{
		if( neighbours == null ) return 0;
		double people = 0;
		for( Cell c : neighbours ) people += c.getPopulation();
		people += getPopulation();
		return people / (neighbours.size() + 1);
	}
	
	public double getCapacityRatio()
	{
		if( !isHabitable() )
		{
			//System.out.println( "Zero capacity because not habitable");
			return 0;
		}
		if( capacity == 0 )
		{
			//System.out.println( "Zero capacity for some reason");
			return 0;
		}
		//return getPopulation() / capacity;
		return (double)getPopulation() / getCapacity();
	}
	
	/**
	 * Removes all the agents from the Cell
	 */
	public void evictResidents()
	{
		//This looks slightly complicated - we can't modify the original agents array
		//while iterating over it, so we create a new array temporarily
		for( Agent a : new ArrayList<Agent>( agents ) ) a.moveOrEmigrate();
		agents.clear();
	}
	
	
	
	/**
	 * Returns true if the cell is surrounded by NEW or OLD development
	 */
	boolean isSurrounded()
	{
		return getDevelopedNeighbours() == neighbours.size();
	}
	
	/**
	 * Return true if the cell has at least one neighbour which is developed (NEW or OLD)
	 * @return
	 */
	public boolean hasNeighbour()
	{
		return getDevelopedNeighbours() > 0;
	}
	
	/**
	 * Returns the number of NEW or OLD neighbours the cell has
	 * @return
	 */
	public int getDevelopedNeighbours()
	{
		int num = 0;
		for( Cell c : neighbours ) if( c.getState().equals( NEW ) || c.getState().equals( OLD ) ) num++;
		return num;
	}
	
	
	/*
	 * Functions after here are unlikely to be useful
	 */
	
	@ScheduledMethod( start = 1, interval=1, priority=6 )
	public void updateState()
	{
		if( nextState != state ) age = 0;
		state = nextState;
	}
	
	protected void preUpdate()
	{
		if( context == null )
			context = (Context<Cell>) ContextUtils.getContext( this );

		if( grid == null )
			grid = (Grid<Cell>) context.getProjection( City.CITY_GRID );

		if( neighbours == null )
		{
			MooreQuery<Cell> query = new MooreQuery<Cell>( grid, this );
			neighbours = new ArrayList<Cell>();
			for ( Cell c : query.query() )
				neighbours.add( c );
		}
	}


	public CellState getState()
	{
		return state;
	}
	
	public String getStateString()
	{
		return state.toString();
	}
	
	public boolean isEmpty() { return CellState.EMPTY.equals( state ); }
	public boolean isAvailable() { return CellState.AVAILABLE.equals( state ); }
	public boolean isNew() { return CellState.NEW.equals( state ); }
	public boolean isOld() { return CellState.OLD.equals( state ); }
	
	public int getCapacity()
	{
		return capacity;
	}

	public int getPopulation()
	{
		return agents.size();
	}
	public int getFreeCapacity()
	{
		if( !isHabitable() ) return 0;
		return max( capacity - agents.size(), 0 );
	}
	
	public boolean add( Agent a )
	{
		if( hasFreeCapacity() )
		{
			agents.add( a );
			return true;
		}
		return false;
	}
	
	public void remove( Agent a )
	{
		agents.remove( a );
	}

	public boolean hasFreeCapacity()
	{
		if( isHabitable() && getFreeCapacity() > 0 ) return true;
		return false;
	}
	
	public boolean isHabitable()
	{
		return ( state.equals( NEW ) || state.equals( OLD  ));
	}
	
	public String getSummary()
	{
		return String.format( "%d(%d) %s:%4.4s\t", getPopulation(), getCapacity(), hasFreeCapacity() ? "Y":"n", getStateString() );
	}
	
	public int getNumMatures()
	{
		int num = 0;
		for( Agent a : agents) if( a instanceof Mature ) num++;
		return num;
	}
	
	public int getNumOlds()
	{
		int num = 0;
		for( Agent a : agents) if( a instanceof Old ) num++;
		return num;
	}
}
