package geoinformatics3.city;

import static geoinformatics3.city.CellState.*;

import java.util.*;

import repast.simphony.context.Context;
import repast.simphony.engine.schedule.*;
import repast.simphony.query.space.grid.*;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.*;
import repast.simphony.util.ContextUtils;

public class CellComplete
{
	public int getAge()
	{
		return age;
	}

	//The current state of the cell
	CellState state = EMPTY;
	//The updated state of the cell, which will be applied once all the cells have calculated their next states
	CellState nextState = EMPTY;
	//The grid on which this cell is located - used to find neighbours etc.
	Grid<CellComplete> grid = null;
	//The cell's Context
	Context<CellComplete> context = null;
	//A list of the cell's neighbours, currently the Moore neighbourhood, i.e. 8 cells
	List<CellComplete> neighbours = null;
	//The number of ticks the cell has spent in the current state
	int age = 0;
	//The probability of empty land with one or more developed neighbours being released for development
	double releaseProbability = 0.4;
	//The probability of released land being built on
	double buildProbability = 0.1;
	//The probability of old development (i.e. age > REDEVELOPMENT_AGE) being redeveloped
	double redevelopmentProbability = 1.05;
	//The age after which cells are considered for redevelopment
	public static final int REDEVELOPMENT_AGE = 30;

	/**
	 * Main update method: cell behaviour goes in here.
	 * 
	 * Hint: to perform probabilistic operations, you can use: if( RandomHelper.nextDouble() < prob ) ...
	 * See also: hasNeighbour() and isSurrounded()
	 */
	@ScheduledMethod(start = 1, interval = 1, priority = 1)
	public void update()
	{
		preUpdate();
		age++;

		switch ( state )
		{
		case EMPTY:
			if( hasNeighbour() && RandomHelper.nextDouble() < releaseProbability ) setState( AVAILABLE );
			break;
		case AVAILABLE:
			if( RandomHelper.nextDouble() < buildProbability ) setState( NEW );
			break;
		case NEW:
			if( isSurrounded() ) setState( OLD );
			break;
		case OLD:
			if( age > REDEVELOPMENT_AGE && RandomHelper.nextDouble() < redevelopmentProbability ) setState( AVAILABLE );
			break;
		}
	}
	
	boolean isSurrounded()
	{
		for( CellComplete c : neighbours ) if( c.getState().equals( EMPTY ) || c.getState().equals( AVAILABLE )) return false;
		return true;
	}
	
	public boolean hasNeighbour()
	{
		return getDevelopedNeighbours() > 0;
	}
	
	public int getDevelopedNeighbours()
	{
		int num = 0;
		for( CellComplete c : neighbours ) if( c.getState().equals( NEW ) || c.getState().equals( OLD ) ) num++;
		return num;
	}
	
	@ScheduledMethod(start=1, interval=1, priority=ScheduleParameters.LAST_PRIORITY )
	public void updateState()
	{
		if( nextState != state ) age = 0;
		state = nextState;
	}
	
	protected void preUpdate()
	{
		if( context == null )
			context = (Context<CellComplete>) ContextUtils.getContext( this );

		if( grid == null )
			grid = (Grid<CellComplete>) context.getProjection( "Grid" );

		if( neighbours == null )
		{
			MooreQuery<CellComplete> query = new MooreQuery<CellComplete>( grid, this );
			neighbours = new ArrayList<CellComplete>();
			for ( CellComplete c : query.query() )
				neighbours.add( c );
		}
	}

	public void setState( CellState state )
	{
		nextState = state;
	}

	public CellState getState()
	{
		return state;
	}
	
	public String getStateString()
	{
		return state.toString();
	}
	
	public boolean isEmpty() { return CellState.EMPTY.equals( state ); }
	public boolean isAvailable() { return CellState.AVAILABLE.equals( state ); }
	public boolean isNew() { return CellState.NEW.equals( state ); }
	public boolean isOld() { return CellState.OLD.equals( state ); }
}
