package geoinformatics4.city;

import geoinformatics4.CombinedModel;

import java.util.*;


import repast.simphony.context.*;
import repast.simphony.context.space.grid.*;
import repast.simphony.dataLoader.ContextBuilder;
import repast.simphony.engine.environment.*;
import repast.simphony.engine.schedule.*;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.*;
import repast.simphony.util.collections.*;
import static geoinformatics4.city.CellState.*;
import static java.lang.Math.*;

public class City
{
	Grid<Cell> grid;
	Context context;
	List<int[]> CENTRAL_POINT = new ArrayList<int[]>();
	List<int[]> LINE = new ArrayList<int[]>();
	List<int[]> CIRCLE = new ArrayList<int[]>();
	List<int[]> STAR = new ArrayList<int[]>();
	List<Cell> cells = new ArrayList<Cell>();
	int size;
	CombinedModel model;
	int totalCapacity = 0;
	int totalFreeCapacity = 0;
	int totalPopulation;
	int newHousesThisTick = 0;
	int unsatisfiedDemand = 0;
	int unsatisfiedDemandCounter = 0;
	
	public static final String CITY_GRID = "CityGrid";
	
	public City( CombinedModel model )
	{
		this.model = model;
	}
	
	public void build( Context context )
	{
		this.context = context;
		init();
		for( int[] coords : CENTRAL_POINT ) setState( NEW, coords );
		context.add( this );
		//RunState.getInstance().getScheduleRegistry().getModelSchedule().schedule( this );
	}
	
	/**
	 * Basic initialisation
	 */
	public void init()
	{
		//get the size of the grid from the environment
		Parameters p = RunEnvironment.getInstance().getParameters();
		size = (Integer) p.getValue( "worldSize" );
		// Create a new 2D grid 
		GridFactory factory =  GridFactoryFinder.createGridFactory( null );
		grid = factory.createGrid( "CityGrid", context, new GridBuilderParameters<Cell>( new StrictBorders(), new SimpleGridAdder<Cell>(), false, size, size ) );

		for( int i = 0; i < size; i++ )
			for( int j = 0; j < size; j++ )
			{
				Cell c = new Cell( model );
				context.add( c );
				grid.getAdder().add( grid, c );
				grid.moveTo( c, i, j );
				c.setLocation( grid.getLocation( c ));
				cells.add( c );
			}
		
		initialiseShapes();
	}
	
	@ScheduledMethod( start = 1.1, interval=1, priority=0 )
	public void update()
	{
		System.out.printf( "NewBuilds: %d, Demand: %d\n", newHousesThisTick, unsatisfiedDemand );
		newHousesThisTick = 0;
		//Update global numbers etc. here
		unsatisfiedDemand = unsatisfiedDemandCounter;
		unsatisfiedDemandCounter = 0;
	}
	
	@ScheduledMethod( start = 1, interval=1, priority=ScheduleParameters.LAST_PRIORITY )
	public void printState()
	{
		//printCity();
		totalCapacity = 0;
		totalFreeCapacity = 0;
		totalPopulation = 0;
		for( Cell c : cells )
		{
			if( c.isHabitable() ) totalCapacity += c.getCapacity();
			totalFreeCapacity += c.getFreeCapacity();
			totalPopulation += c.getPopulation();
		}
		
	}
	
	/**
	 * Sets the state of one of the cells in the grid
	 * @param state
	 * @param pos
	 */
	void setState(  CellState state, int... pos )
	{
		Cell c = grid.getObjectAt( pos );
		c.setState( state );
		c.updateState();
		
	}
	
	/**
	 * Initialises the possible starting shapes
	 */
	void initialiseShapes()
	{
		int minDim = min( size, size );
		int cx = size/2;
		int cy = size/2;
		
		//Central point is just one coordinate
		CENTRAL_POINT.add( new int[]{ cx, cy } );
		//Horizontal line from 1/4 to 3/4 of the width
		for( int i = size/4; i < 3*size/4; i++ )
			LINE.add( new int[]{i, cy } );
		
		//All points within 1/6 of the centre
		for( int i = 0; i < size; i++ )
			for( int j = 0; j < size; j++ )
				if( ( cx - i ) * (cx-i) + (cy-j)* (cy-j) < minDim / 6 )
					CIRCLE.add( new int[] { i, j } );
		
		//Points on diagonals
		STAR.add( new int[]{ cx, cy } );
		for( int i = 1; i < minDim/6; i++ )
		{
			STAR.add( new int[]{ cx-i, cy-i } );
			STAR.add( new int[]{ cx+i, cy-i } );
			STAR.add( new int[]{ cx+i, cy+i } );
			STAR.add( new int[]{ cx-i, cy+i } );
		}
		
	}
	
	public Iterable<Cell> getCells()
	{
		return cells;
	}
	
	public Iterable<Cell> getRandomCells( int number )
	{
		return context.getRandomObjects( Cell.class, number );
	}
	
	public Iterable<Cell> getRandomAvailableCells( int max )
	{
		List<Cell> available = new ArrayList<Cell>(cells.size());
		for( Cell c : cells ) 
			if( c.hasFreeCapacity() ) available.add( c );
		while( available.size() > max )
			available.remove( RandomHelper.nextIntFromTo( 0, available.size()-1 ) );
		Collections.shuffle( available );
		return available;
	}
	
	void printCity()
	{
		System.out.println( "Printing city: " + this );
		for( int i = size-1; i >= 0; i-- )
		{
			System.out.println();
			for( int j = 0; j < size; j++ )
			{
				Cell c = grid.getObjectAt( j,i );
				System.out.printf( c.getSummary() );
			}
		}
			System.out.println();
	}
	
	public void reinitialiseCellCapacities()
	{
		for( Cell c : getCells() ) c.capacity = Cell.INITIAL_CAPACITY;
	}

	public int getTotalCapacity()
	{
		return totalCapacity;
	}

	public int getTotalFreeCapacity()
	{
		return totalFreeCapacity;
	}

	public int getTotalPopulation()
	{
		return totalPopulation;
	}

	public int getNewHousesThisTick()
	{
		return newHousesThisTick;
	}
	
	public void addHouses( int number )
	{
		newHousesThisTick += number;
	}
	
	public void unsatisfiedDemmand()
	{
		unsatisfiedDemandCounter++;
	}

	public int getUnsatisfiedDemand()
	{
		return unsatisfiedDemand;
	}

	public Grid<Cell> getGrid()
	{
		return grid;
	}

	public int getSize()
	{
		return size;
	}
}
