package gi.city;

import gi.agents.*;
import gi.CombinedModel;
import static gi.city.CellState.*;

import java.util.*;
import static java.lang.Math.*;

import repast.simphony.context.Context;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.query.space.grid.MooreQuery;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.grid.*;
import repast.simphony.util.ContextUtils;

/**
 * You shouldn't need to edit Cell for this exercise, but some of the functions
 * will prove useful!
 * @author dmrust
 *
 */
public class Cell
{
	CombinedModel model;
	City city;
	GridPoint location;

	//The current state of the cell
	CellState state = EMPTY;
	//The updated state of the cell, which will be applied once all the cells have calculated their next states
	CellState nextState = EMPTY;
	//The grid on which this cell is located - used to find neighbours etc.
	Grid<Cell> grid = null;
	//The cell's Context
	Context<Cell> context = null;
	//A list of the cell's neighbours, currently the Moore neighbourhood, i.e. 8 cells
	List<Cell> neighbours = null;
	//The number of ticks the cell has spent in the current state
	int age = 0;
	//The probability of empty land with one or more developed neighbours being released for development
	public static double RELEASE_PROBABILITY = 0.1;
	//The probability of old development (i.e. age > REDEVELOPMENT_AGE) being redeveloped
	public static double REDEVELOPMENT_RATE = 0.2;
	//The age after which cells are considered for redevelopment
	public static int REDEVELOPMENT_AGE = 30;
	//The proportion of housing to allow to be surplus
	public static double SURPLUS_FACTOR = 0.1;
	
	
	//The current capacity of the cell
	public static int MAX_CAPACITY = 1;
	public static int INITIAL_CAPACITY = 1;
	public static int CAPACITY_INCREMENT = 1;
	public static int CURRENT_MAXIMUM_CAPACITY = INITIAL_CAPACITY;
	int capacity = INITIAL_CAPACITY;
	//The agents who are living in the cell
	Set<Agent> agents = new HashSet<Agent>(10);
	
	double distanceToRiver = Double.MAX_VALUE;
	String feature;
	
	public Cell( CombinedModel model )
	{
		this.model = model;
		this.city = model.getCity();
	}
	
	/**
	 * These are basic values - you can change them if you want, and you should certainly
	 * alter the EMPTY case to take into account closeness to the river.
	 * 
	 */
	@ScheduledMethod( start = 1, interval=1, priority=10 )
	public void update()
	{
		preUpdate();
		age++;

		switch ( state )
		{
		case EMPTY:
			if( hasNeighbour() && RandomHelper.nextDouble() < RELEASE_PROBABILITY ) setState( AVAILABLE );
			break;
		case AVAILABLE:
			double prob = 1 - city.getTotalCapacity() / (city.getTotalPopulation()*(1+SURPLUS_FACTOR));
			if( RandomHelper.nextDouble() < prob ) build();

			break;
		case NEW:
			if( getPopulation() > 0 ) setState(OLD);
			break;
		case OLD:
			if( age >= REDEVELOPMENT_AGE && capacity < MAX_CAPACITY )
				if( RandomHelper.nextDouble() < getNeighbourhoodCapacityRatio() * REDEVELOPMENT_RATE ) redevelop();
			break;
		}
	}
	
	/**
	 * Use this function to redevelop cells; it evicts all the residents, increases the capacity
	 * and rebuilds
	 */
	void redevelop()
	{
		evictResidents();
		model.getCity().addHouses( -capacity );
		incrementCapacity();
		build();
	}
	
	/**
	 * Use this function to transition from AVAILABLE to NEW - it updates the city about the
	 * current housing capacity
	 */
	void build()
	{
		setState( NEW );
		model.getCity().addHouses( capacity );
	}
	
	public void incrementCapacity()
	{
		capacity += CAPACITY_INCREMENT;
		if( capacity > MAX_CAPACITY ) capacity = MAX_CAPACITY;
		if( capacity > CURRENT_MAXIMUM_CAPACITY ) CURRENT_MAXIMUM_CAPACITY = capacity;
	}
	
	/*
	 * Functions after here are useful, but should not need changing
	 */
	
	/**
	 * Use this to set the state of the cell to a new state (e.g. setState(NEW) )
	 * @param state
	 */
	public void setState( CellState state )
	{
		nextState = state;
	}
	
	/**
	 * Returns the average ratio of population to capcity in the neighbourhood
	 * @return
	 */
	public double getNeighbourhoodCapacityRatio()
	{
		preUpdate();
		if( neighbours == null ) return 0;
		double fullness = 0;
		for( Cell c : neighbours ) fullness += c.getCapacityRatio();
		System.out.println( "Surrounding cells: " + fullness + " (" + neighbours.size() + ")" );
		System.out.println( "My fullness: " + getCapacityRatio() );
		fullness += getCapacityRatio();
		return fullness / (neighbours.size() + 1);
	}
	
	public double getNeighbourhoodDensity()
	{
		if( neighbours == null ) preUpdate();
		return getNeighbourhoodPopulation() / (neighbours.size() + 1);
	}
	
	public int getNeighbourhoodPopulation()
	{
		if( neighbours == null ) return 0;
		int people = 0;
		for( Cell c : neighbours ) people += c.getPopulation();
		people += getPopulation();
		return people;
	}
	
	public double getCapacityRatio()
	{
		if( !isHabitable() )
		{
			//System.out.println( "Zero capacity because not habitable");
			return 0;
		}
		if( capacity == 0 )
		{
			//System.out.println( "Zero capacity for some reason");
			return 0;
		}
		//return getPopulation() / capacity;
		return (double)getPopulation() / getCapacity();
	}
	
	/**
	 * Removes all the agents from the Cell
	 */
	public void evictResidents()
	{
		//This looks slightly complicated - we can't modify the original agents array
		//while iterating over it, so we create a new array temporarily
		for( Agent a : new ArrayList<Agent>( agents ) ) a.moveOrEmigrate();
		agents.clear();
	}
	
	
	
	/**
	 * Returns true if the cell is surrounded by NEW or OLD development
	 */
	boolean isSurrounded()
	{
		return getDevelopedNeighbours() == neighbours.size();
	}
	
	/**
	 * Return true if the cell has at least one neighbour which is developed (NEW or OLD)
	 * @return
	 */
	public boolean hasNeighbour()
	{
		return getDevelopedNeighbours() > 0;
	}
	
	/**
	 * Returns the number of NEW or OLD neighbours the cell has
	 * @return
	 */
	public int getDevelopedNeighbours()
	{
		int num = 0;
		for( Cell c : neighbours ) if( c.getState().equals( NEW ) || c.getState().equals( OLD ) ) num++;
		return num;
	}
	
	
	/*
	 * Functions after here are unlikely to be useful
	 */
	
	@ScheduledMethod( start = 1, interval=1, priority=6 )
	public void updateState()
	{
		if( nextState != state ) age = 0;
		state = nextState;
	}
	
	protected void preUpdate()
	{
		if( context == null )
			context = (Context<Cell>) ContextUtils.getContext( this );

		if( grid == null )
			grid = (Grid<Cell>) context.getProjection( City.CITY_GRID );

		if( neighbours == null )
		{
			MooreQuery<Cell> query = new MooreQuery<Cell>( grid, this, 1, 1 );
			//VNQuery<Cell> query = new VNQuery<Cell>( grid, this, 1, 1 );
			neighbours = new ArrayList<Cell>();
			for ( Cell c : query.query() )
			{
				//Not sure why we're getting ourselves in the query results, but we are!
				if( c != this ) neighbours.add( c );
			}
		}
	}


	public CellState getState()
	{
		return state;
	}
	
	public String getStateString()
	{
		return state.toString();
	}
	
	public boolean isEmpty() { return CellState.EMPTY.equals( state ); }
	public boolean isAvailable() { return CellState.AVAILABLE.equals( state ); }
	public boolean isNew() { return CellState.NEW.equals( state ); }
	public boolean isOld() { return CellState.OLD.equals( state ); }
	
	public int getCapacity()
	{
		return capacity;
	}

	public int getPopulation()
	{
		return agents.size();
	}
	public int getFreeCapacity()
	{
		if( !isHabitable() ) return 0;
		return max( capacity - agents.size(), 0 );
	}
	
	public boolean add( Agent a )
	{
		if( hasFreeCapacity() )
		{
			agents.add( a );
			return true;
		}
		return false;
	}
	
	public void remove( Agent a )
	{
		agents.remove( a );
	}

	public boolean hasFreeCapacity()
	{
		if( isHabitable() && getFreeCapacity() > 0 ) return true;
		return false;
	}
	
	public boolean isHabitable()
	{
		return ( state.equals( NEW ) || state.equals( OLD  ));
	}
	
	public String getSummary()
	{
		return String.format( "%d(%d) %s:%4.4s\t", getPopulation(), getCapacity(), hasFreeCapacity() ? "Y":"n", getStateString() );
	}
	
	public int getNumMatures()
	{
		int num = 0;
		for( Agent a : agents) if( a instanceof Mature ) num++;
		return num;
	}
	
	public int getNumOlds()
	{
		int num = 0;
		for( Agent a : agents) if( a instanceof Old ) num++;
		return num;
	}

	public Set<Agent> getAgents()
	{
		return agents;
	}
	
	public int countAgentsByType( AgentType type )
	{
		int i = 0;
		for( Agent a : agents ) if( a.getType().equals( type )) i++;
		return i;
	}
	
	public AgentType getMajorityType()
	{
		if( agents.size() == 0 ) return null;
		int curMax = 0;
		AgentType curType = null;
		for( AgentType t : AgentType.values() )
		{
			int num = countAgentsByType( t );
			if( num > curMax )
			{
				curMax = num;
				curType = t;
			}
		}
		return curType;
	}
	
	public Map<AgentType,Integer> getNeighbourhoodTypes()
	{
		preUpdate();
		Map<AgentType,Integer> agents = new HashMap<AgentType, Integer>();
		for( AgentType t : AgentType.values() ) 
		{
			int num = countAgentsByType( t );
			for( Cell c : neighbours ) num += c.countAgentsByType( t );
			agents.put( t, num );
		}
		return agents;
	}
	
	public double getProportionOfTypeInNeighbourhood( AgentType t )
	{
		int like = getNeighbourhoodTypes().get( t );
		int all = getNeighbourhoodPopulation();
		return (double)like / all;	
	}

	public void setLocation( GridPoint point )
	{
		location = point;
	}
	
	public GridPoint getLocation()
	{
		return location;
	}
	public int getAge()
	{
		return age;
	}

	public double getDistanceToRiver()
	{
		return distanceToRiver;
	}

	public void setDistanceToRiver( double distanceToRiver )
	{
		this.distanceToRiver = distanceToRiver;
	}

	public String getFeature()
	{
		return feature;
	}

	public void setFeature( String feature )
	{
		this.feature = feature;
	}
	
	public List<Cell> getNeighbours()
	{
		preUpdate();
		return neighbours;
	}
}
